#! /usr/bin/env python
""" Parrot

48 hour game development contest

Language: Python 2.2:  http://www.python.org
Library: PyGame 1.2: http://www.pygame.org
"""

# pygame basics
import pygame, pygame.image, pygame.transform, pygame.sprite, pygame.draw
from pygame.locals import *

# other libraries
import random, os.path, sys, array, time, string

# configuration parameters:
FrameRate = 30

#ScreenSize = (640, 480)
ScreenSize = (800, 600)
#ScreenSize = (1024, 768)
GridSize = 7

#WinStyle = FULLSCREEN
WinStyle = HWPALETTE

GridOrigin = (ScreenSize[0]/2, 200)

# Globals:
g_frame = 0
g_level = 1
g_grid = None
g_theparrot = None
g_shapes_image = None
g_sprites = None
g_static_image = None
g_gameover = 0

def GameOver():
    global g_gameover
    g_gameover = 1

def Level(level=None):
    global g_level, g_grid
    if level==None:
        level = g_level + 1

    g_level = level
    FrameRate = 25+level*2

    f = open('levels.map')
    lines = f.readlines()
    f.close()
    
    factory_dict = {
        'p': (lambda abpos: Parrot(g_sprites, g_shapes_image, ab_to_xy(abpos),0)),
        'c': (lambda abpos: Camel(g_sprites, g_shapes_image, ab_to_xy(abpos),0)),
        's': (lambda abpos: Snake(g_sprites, g_shapes_image, ab_to_xy(abpos),0)),
        }

    g_grid = Grid(g_shapes_image, GridSize, lines[(level-1)*14:], factory_dict)
    
    # initial drawing
    g_static_image.blit(g_background_image, (0, 0))
    g_grid.draw(g_static_image)
    g_surface.blit(g_static_image, (0, 0))
    pygame.display.update()
    if level>1:
        g_theparrot.score += 100*Snake.count
    g_theparrot.score_dirty = 1
    


class RenderZSort(pygame.sprite.RenderUpdates):
    def draw(self, surface):
        "draw all sprites onto a surface in z order (lowest z first)"
        spritedict = self.spritedict
        items = spritedict.items()
        items.sort(lambda a,b: cmp(a[0].z,b[0].z))
        surface_blit = surface.blit
        dirty = self.lostsprites
        self.lostsprites = []
        dirty_append = dirty.append

        for s, r in items:
            newrect = surface_blit(s.image, s.rect)
            s.post_draw(surface)
            if r is not 0:
                dirty_append(newrect.union(r))
            else:
                dirty_append(newrect)
            spritedict[s] = newrect
        return dirty

# Text graphics
#
# Text effects are implemented by editing a photoshop file with a text layer
# containing a row of fixed width printable characters:  text.psd
#
# The text layer is pre-rendered as 24-bit png in photoshop to desired specifications:
# color, contrast, bevel shadow and oter effects, etc.
#
# The text.psd file is included
# !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~ 

class BlitFont:
    def __init__(self, background, file, (width,height), print_width):
        'Fixed width font from bitmap'
        self.background = background
        self.image = pygame.image.load(file).convert_alpha()
        self.images = {}
        self.width = width
        self.height = height
        self.print_width = print_width

        for j in range(95):
            surface = self.image.subsurface((j*width,0,width,height))
            self.images[chr(32+j)] = surface

    def blit_text(self, surface, s, (x,y)):
        "blits s to surface at (x,y) and returns an update rectangle"
        r = (x,y,self.width*len(s), self.height)
        surface.blit(self.background, (x,y), r)
        for i in range(len(s)):
            surface.blit(self.images[s[i]], (x+i*self.print_width,y))
        return r

# Each sprite has a png file containing its frames arranged horizontally
# Like the BlitFont class, 

# directions:
#
# 3   0
#  \ /
#  / \
# 2   1

def ab_to_xy((a,b)):
    x0,y0 = GridOrigin
    x = x0+(b-a)*64
    y = y0+(b+a)*32
    return (x,y)

def xy_to_ab((x,y)):
    x0,y0 = GridOrigin
    a = int(((y-y0+32) - (x-x0)/2)/64)
    b = int(((y-y0+32) + (x-x0)/2)/64)
    exact = ((x,y) == ab_to_xy((a,b)))
    return ((a,b), exact)
    

class Object:
    def __init__(self, image, origin, size, root, position):
        w,h = size
        x0,y0 = origin
        self.image = image.subsurface((x0,y0,w,h))
        self.root = root
        self.position = position

    def draw(self, surface):
        (x,y) = self.position
        (rx,ry) = self.root
        surface.blit(self.image, (x-rx,y-ry))

class Wall:
    def __init__(self, image, walltype, transtype, abpos):
        w = h = 128
        (x0,y0) = (walltype*128, 384)
        self.image = image.subsurface((x0,y0,w,h))
        if transtype>=8:
            transtype -= 8
            (x0,y0) = (transtype*128, 640)
        else:
            (x0,y0) = (transtype*128, 512)
        self.timage = image.subsurface((x0,y0,w,h))

        if walltype<4:
            self.root = (96,96)
        else:
            self.root = (32,96)
            
        self.position = ab_to_xy(abpos)

    def draw(self, surface):
        (x,y) = self.position
        (rx,ry) = self.root
        
        surface.blit(self.image, (x-rx,y-ry))

    def draw_transparent(self, surface):
        (x,y) = self.position
        (rx,ry) = self.root
        
        return surface.blit(self.timage, (x-rx,y-ry))

class Grid:
    def __init__(self, image, size, lines, factory):
        self.size = size
        self.cells = []
        lines.append('  '*(size+1))
        lines.append('  '*(size+1))
        for a in range(size):
            row = []
            lines[a*2] = lines[a*2][:-1] + '  '*(size+1)
            lines[a*2+1] = lines[a*2+1][:-1] + '  '*(size+1)
            for b in range(size):
                cell = Cell()
                row.append(cell)

                if lines[a*2+1][b*2] != ' ':
                    cell.wa = 1
                if lines[a*2][b*2+1] != ' ':
                    cell.wb = 1

                c = lines[a*2+1][b*2+1]
                if  c != ' ':
                    if c in factory:
                        factory[c]((a,b))
                    else:
                        print "unknown character '%s' at %s" % (c, `(a,b)`)
                
            self.cells.append(row)

        for a in range(size):
            for b in range(size):
                cell = self.cells[a][b]
                if cell.wa:
                    t = 0
                    if cell.wb:
                        t+=1
                    if a+1<size and self.cells[a+1][b].wb:
                        t+=2
                    ts = t

                    if a+1<size and self.cells[a+1][b].wa and b>0 and self.cells[a+1][b-1].wb:
                        ts += 8

                    if a>0 and self.cells[a-1][b].wa:
                        t |= 2

                    if b>0 and self.cells[a][b-1].wb:
                        t |= 2

                    cell.wa = Wall(image, t, ts, (a,b))
                if cell.wb:
                    t = 4
                    if cell.wa:
                        t+=1
                    if b+1<size and self.cells[a][b+1].wa:
                        t+=2
                    ts = t

                    if a>0 and self.cells[a-1][b+1].wa:
                        t |= 2
                    else:
                        t = 4
                    
                    cell.wb = Wall(image, t, ts, (a,b))

    def draw(self, surface):
        for a in range(self.size):
            for b in range(self.size):
                cell = self.cells[a][b]
                if cell.wa:
                    cell.wa.draw(surface)
                if cell.wb:
                    cell.wb.draw(surface)

    def draw_transparent_cell(self, surface, (a,b)):
        update_list = []
        cell = self.cells[a][b]
        if self.cells[a+1][b].wb:
            update_list.append(self.cells[a+1][b].wb.draw_transparent(surface))
        if self.cells[a][b+1].wa:
            update_list.append(self.cells[a][b+1].wa.draw_transparent(surface))
        if self.cells[a+1][b+1].wa:
            update_list.append(self.cells[a+1][b+1].wa.draw_transparent(surface))
        if self.cells[a+1][b+1].wb:
            update_list.append(self.cells[a+1][b+1].wb.draw_transparent(surface))
        return update_list

    def get_options(self, (a,b)):
        dlist = []
        cells = self.cells
        if not cells[a][b].wb:
            dlist.append(0)
        if not (b+1>=self.size or cells[a][b+1].wa):
            dlist.append(1)
        if not (a+1>=self.size or cells[a+1][b].wb):
            dlist.append(2)
        if not cells[a][b].wa:
            dlist.append(3)
        return dlist

class Cell:
    def __init__(self):
        self.wa = self.wb = None


class Creature(pygame.sprite.Sprite):
    def __init__(self, container, image, origin, position, d):
        pygame.sprite.Sprite.__init__(self, container)
        self.container = container
        self.main_image = image
        self.root = (32, 48)
        self.images = []
        self.dvector = {-1:(0,0), 0:(2,-1), 1:(2,1), 2:(-2,1), 3:(-2,-1)}
        x0,y0 = origin
        self.d = d
        self.target = None

        for d in range(4):
            for i in range(4):
                surface = image.subsurface((x0+(d*4+i)*64,y0,64,64))
                self.images.append(surface)

        x,y = self.position = position
        self.z = y
        self.update()

    def get_ab(self):
        return xy_to_ab(self.position)[0]

    def update(self):
        (x,y) = self.position
        (rx,ry) = self.root
        (a,b),exact = xy_to_ab((x,y))
        if exact and g_grid:
            dirs = g_grid.get_options((a,b))
            d = self.choose((a,b), dirs)
            if d^self.d==2:
                d = self.choose((a,b), dirs)
            self.d = d

        if g_grid:
            (dx,dy) = self.dvector[self.d]
            x+=dx; y+=dy
            self.position = (x,y)
        self.z = y
        self.image = self.images[self.d*4 + ((x/2)%4)]
        self.rect = pygame.Rect(x-rx,y-ry, 64,192)

    def choose(self, (a,b), dirs):
        return random.choice(dirs)

    def post_draw(self, surface):
        g_grid.draw_transparent_cell(surface, xy_to_ab(self.position)[0])

class Parrot(Creature):
    def __init__(self, container, image, position, d):
        global g_theparrot
        g_theparrot = self
        self.stun = 0
        self.score = 0
        self.score_dirty = 1
        Creature.__init__(self, container, image, (0,128), position, d)

    def update(self):
        Creature.update(self)
        container = self.groups()[0]
        for obj in pygame.sprite.spritecollide(self, container, 0):
            if obj.get_ab() != self.get_ab():
                continue
            if obj.__class__ == Camel:
                if not self.stun:
                    Mutant(container, obj)
                    obj.kill()
                    g_theparrot.score += 10*(Snake.count+1)
                    g_theparrot.score_dirty = 1
            if obj.__class__ == CamelSpit:
                obj.kill()
                self.stun = 20

    def choose(self, (a,b), dirs):
        if self.stun:
            self.stun -= 1
            return -1

        if self.target == (a,b):
            self.target = None

        if self.target:
            ta,tb = self.target
            if abs(ta-a) > abs(tb-b):
                if ta>a and (2 in dirs):
                    return 2
                if ta<a and (0 in dirs):
                    return 0
                if tb>b and (1 in dirs):
                    return 1
                if tb<b and (3 in dirs):
                    return 3
            else:
                if tb>b and (1 in dirs):
                    return 1
                if tb<b and (3 in dirs):
                    return 3
                if ta>a and (2 in dirs):
                    return 2
                if ta<a and (0 in dirs):
                    return 0

        return -1
        
            
        return random.choice(dirs)
#            self.score += 10*random.randint(2,5)
#            self.score_dirty = 1

class Snake(Creature):
    count = 0

    def __init__(self, container, image, position, d):
        Snake.count += 1
        Creature.__init__(self, container, image, (0,64), position, d)
        
    def kill(self):
        global g_gameover
        Snake.count -= 1
        Creature.kill(self)
        if Snake.count == 0:
            g_gameover = 1

class Camel(Creature):
    count = 0

    def __init__(self, container, image, position, d):
        Camel.count += 1
        Creature.__init__(self, container, image, (0,0), position, d)

    def kill(self):
        Camel.count -= 1
        Creature.kill(self)
        if Camel.count == 0:
            Level()

    def update(self):
        Creature.update(self)
        if random.random()>0.99-g_level*0.01:
            CamelSpit(self.container, self)

class CamelSpit(Creature):
    def __init__(self, container, camel):
        self.age = 0
        Creature.__init__(self, container, camel.main_image, (0,256), camel.position, camel.d)

    def update(self):
        Creature.update(self)
        Creature.update(self)
        for obj in pygame.sprite.spritecollide(self, self.container, 0):
            if obj.get_ab() != self.get_ab():
                continue
            if obj.__class__ == Snake:
                SnakeMelty(self.container, obj)
                obj.kill()
        x,y = self.position
        self.image = self.images[self.d*4 + ((x/8)%4)]

    def choose(self, (a,b), dirs):
        if self.d in dirs:
            return self.d
        else:
            self.kill()
            return -1
    
class Mutant(Creature):
    def __init__(self, container, camel):
        self.container = container
        self.age = 0
        Creature.__init__(self, container, camel.main_image, (0,192), camel.position, camel.d)

    def update(self):
        (x,y) = self.position
        (rx,ry) = self.root
        self.z = y
        self.image = self.images[self.d*4 + self.age/16]
        self.rect = pygame.Rect(x-rx,y-ry, 64,192)
        self.age += 1
        
        if self.age>=64:
            Snake(self.container, self.main_image, self.position, self.d)
            self.kill()
            del self.container
            
class SnakeMelty(Creature):
    def __init__(self, container, snake):
        self.container = container
        self.age = 0
        Creature.__init__(self, container, snake.main_image, (0,320), snake.position, snake.d)

    def update(self):
        (x,y) = self.position
        (rx,ry) = self.root
        self.z = y
        self.image = self.images[self.d*4 + self.age/8]
        self.rect = pygame.Rect(x-rx,y-ry, 64,192)
        self.age += 1
        
        if self.age>=32:
            self.kill()
            del self.container

def main(winstyle = 0):
    global g_frame, g_grid, g_shapes_image, g_sprites, g_static_image, g_background_image, g_surface

    # Initialize pygame
    pygame.init()

    pygame.display.set_mode(ScreenSize, WinStyle, 16)
    pygame.display.set_caption('Pray for World Peace!')

    g_surface = pygame.display.get_surface()

    # background
    g_background_image = pygame.image.load('graphics\\background.png').convert()
    g_static_image = g_background_image.convert()

    # sprites and the grid
    g_shapes_image = pygame.image.load('graphics/creatures.png').convert_alpha()
    g_sprites = RenderZSort()

    # score graphics
    font = BlitFont(g_background_image, 'graphics/text.png', (30,40), 24)

    Level(1)

    # frame timing
    clock = pygame.time.Clock()
    t0 = time.time()

    pygame.event.set_blocked([MOUSEMOTION])

    quit = 0
    while not quit:
        pygame.event.pump()
        update_list = []

        for e in pygame.event.get():
            if e.type == QUIT:
                quit = 1
        
            if e.type == KEYDOWN:
                if e.key == K_ESCAPE:
                    quit = 1
 
            if e.type == MOUSEBUTTONDOWN:
                (a,b), exact =  xy_to_ab(e.pos)
                g_theparrot.target = (a,b)

        # clear sprites
        g_sprites.clear(g_surface, g_static_image)

        # draw the score
        if g_theparrot.score_dirty:
            r = font.blit_text(g_static_image, '%7d' % g_theparrot.score, (10,10))
            g_surface.blit(g_static_image, r[:2], r)
            update_list.append(r)
            r = font.blit_text(g_static_image, 'Level %d' % g_level, (600,10))
            g_surface.blit(g_static_image, r[:2], r)
            update_list.append(r)
            g_theparrot.score_dirty = 0

        if g_gameover:
            r = font.blit_text(g_static_image, 'Game Over', (280,10))
            g_surface.blit(g_static_image, r[:2], r)
            update_list.append(r)
        
        # draw sprites
        if not g_gameover:
            g_sprites.update()
        update_list += g_sprites.draw(g_surface)
        #update_list += g_grid.draw_transparent(g_surface)

        # process frame
        pygame.display.update(update_list)
        clock.tick(FrameRate)
        g_frame = g_frame + 1

if __name__ == '__main__': main()

